package server

import (
	"context"
	"errors"
	"github.com/odysseus/nodemanager/utils"
	omanager "github.com/odysseus/odysseus-protocol/gen/proto/go/nodemanager/v1"
	log "github.com/sirupsen/logrus"
	"strconv"
	"strings"
)

var (
	ErrParams    = errors.New("invalid request")
	ErrUnSupport = errors.New("unsupport feature")
)

type NodeManagerService struct {
	node *Node
	quit chan struct{}
	omanager.UnimplementedNodeManagerServiceServer
}

func (n *NodeManagerService) ManagerList(ctx context.Context, request *omanager.ManagerListRequest) (*omanager.ManagerListResponse, error) {
	list := n.node.register.GetNodeManagerList(nil)
	res := new(omanager.ManagerListResponse)
	res.Managers = make([]*omanager.NodeManagerInfo, 0, len(list))
	for _, v := range list {
		res.Managers = append(res.Managers, &omanager.NodeManagerInfo{
			Publickey: v.Pubkey,
			Endpoint:  v.Endpoint,
		})
	}
	return res, nil
}

func (n *NodeManagerService) RegisterWorker(client omanager.NodeManagerService_RegisterWorkerServer) error {
	//return n.node.wm.handleNewDial(client)

	uuid := utils.GetSnowflakeId()
	log.WithField("uuid", uuid).Debug("new worker register")

	worker, err := n.node.wm.AddNewWorker(uuid, client)
	if err != nil {
		log.WithField("uuid", uuid).WithError(err).Error("add new worker failed")
		return err
	}
	return n.node.wm.manageWorker(worker)
}

func (n *NodeManagerService) DispatchTask(ctx context.Context, request *omanager.DispatchTaskRequest) (*omanager.DispatchTaskResponse, error) {
	if request == nil {
		return nil, errors.New("invalid request")
	}
	mids := strings.Split(request.Miner, "_")
	if len(mids) != 2 {
		return nil, errors.New("invalid miner")
	}
	workerAddr := mids[0]
	worker := n.node.wm.GetWorkerByAddr(workerAddr)
	if worker == nil {
		return nil, errors.New("worker not found")
	}
	if worker.online == false {
		return nil, errors.New("worker offline")
	}
	{
		nonceds := strings.Split(mids[1], ":")
		nonce, _ := strconv.ParseInt(nonceds[0], 10, 64)
		if nonce < int64(worker.nonce) {
			return nil, errors.New("expired worker nonce")
		}
	}

	dtask := newDispatchTask(worker, request.TaskData)

	worker.taskCh <- dtask

	// wait task send to worker.
	if err, _ := <-dtask.errCh; err != nil {
		return nil, err
	}

	res := new(omanager.DispatchTaskResponse)
	res.TaskId = request.TaskData.TaskId
	res.Miner = workerAddr
	return res, nil
}
