package server

import (
	"errors"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/odysseus/nodemanager/utils"
	odysseus "github.com/odysseus/odysseus-protocol/gen/proto/go/base/v1"
	omanager "github.com/odysseus/odysseus-protocol/gen/proto/go/nodemanager/v1"
	log "github.com/sirupsen/logrus"
	"math/big"
	"time"
)

var (
	ErrExecuteTimeout = errors.New("execute timeout")
	ErrExecuteFailed  = errors.New("execute failed")
)

func (wm *WorkerManager) taskResult(worker *Worker, task *odysseus.TaskContent, result *omanager.SubmitTaskResult) (*odysseus.TaskProof, error) {
	switch task.TaskKind {
	case odysseus.TaskKind_ComputeTask:
		return wm.computeTaskResult(worker, task, result)
	case odysseus.TaskKind_StandardTask:
		return wm.standardTaskResult(worker, task, result)
	}

	return nil, errors.New("unsupport task kind")
}

func (wm *WorkerManager) computeTaskResult(worker *Worker, task *odysseus.TaskContent, result *omanager.SubmitTaskResult) (*odysseus.TaskProof, error) {
	if worker.info.nodeInfo == nil {
		return nil, errors.New("unknown worker node info")
	}
	//log.WithFields(log.Fields{
	//	"task-id":               task.TaskId,
	//	"task-type":             task.TaskType,
	//	"task-kind":             task.TaskKind,
	//	"task-success":          result.IsSuccessed,
	//	"task-execute-duration": result.TaskExecuteDuration,
	//}).Debug("got task result")

	if result.IsSuccessed == false {
		taskResponse := &odysseus.TaskResponse{
			TaskId:           task.TaskId,
			TaskResultHeader: result.TaskResultHeader,
			TaskResultBody:   result.TaskResultBody,
			TaskUid:          task.TaskUid,
			TaskFee:          task.TaskFee,
			TaskIsSucceed:    false,
			TaskResultCode:   result.TaskExecuteCode,
			TaskExecuteError: result.TaskExecuteError,
			TaskError:        ErrExecuteFailed.Error(),
		}
		receipt := wm.makeReceipt(worker, task, result, ErrExecuteFailed)
		wm.node.PostResult(receipt)
		go wm.doCallback(task.TaskCallback, taskResponse)
		return nil, nil
	} else {
		taskResponse := &odysseus.TaskResponse{
			TaskId:           task.TaskId,
			TaskResultHeader: result.TaskResultHeader,
			TaskResultBody:   result.TaskResultBody,
			TaskUid:          task.TaskUid,
			TaskFee:          task.TaskFee,
			TaskIsSucceed:    true,
			TaskResultCode:   result.TaskExecuteCode,
			TaskExecuteError: result.TaskExecuteError,
			TaskError:        "",
		}
		go wm.doCallback(task.TaskCallback, taskResponse)
	}
	paramHash := crypto.Keccak256Hash(task.TaskParam)
	resultHash := crypto.Keccak256Hash(result.TaskResultBody)
	{
		// verify container_signature and miner_signature
		// container_signature = sign(hash(task_id+hash(task_param)+hash(task_result)))
		verified := utils.VerifySignature(utils.CombineBytes([]byte(result.TaskId), paramHash[:], resultHash[:]),
			result.ContainerSignature, task.ContainerPubkey)
		log.WithFields(log.Fields{
			"task-id":                  task.TaskId,
			"containerSignatureVerify": verified,
			"taskkind":                 task.TaskKind,
		}).Debug("container signature verify")
		if !verified {
			// todo: handle signature verify failed
		}
	}
	{
		// verify miner_signature
		// miner_signature = sign(hash((task_id+hash(task_param)+hash(task_result)))
		verified := utils.VerifySignature(utils.CombineBytes([]byte(result.TaskId), paramHash[:], resultHash[:]),
			result.MinerSignature, utils.FromHex(worker.info.nodeInfo.MinerPubkey))
		log.WithFields(log.Fields{
			"task-id":              task.TaskId,
			"minerSignatureVerify": verified,
			"taskkind":             task.TaskKind,
		}).Debug("miner signature verify")
		if !verified {
			// todo: handle signature verify failed
		}
	}

	receipt := wm.makeReceipt(worker, task, result, Succeed)
	wm.node.PostResult(receipt)

	//manager_signature = sign(hash((task_id+hash(task_param)+hash(task_result)+container_signature+miner_signature+workload+time))
	now := time.Now().Unix()
	dataHash := crypto.Keccak256Hash(utils.CombineBytes([]byte(result.TaskId), paramHash[:], resultHash[:],
		worker.ProfitAccount().Bytes(), worker.WorkerAccount().Bytes(), result.ContainerSignature, result.MinerSignature, big.NewInt(int64(task.TaskWorkload)).Bytes()),
		big.NewInt(now).Bytes())

	signature, err := wm.node.Sign(dataHash[:])
	if err != nil {
		log.WithError(err).Error("sign result failed")
		return nil, err
	}

	proof := wm.makeTaskProof(worker, task, taskProof{
		paramHash:          paramHash.Bytes(),
		resultHash:         resultHash.Bytes(),
		finishTime:         uint64(now),
		nmSignature:        signature,
		containerSignature: result.ContainerSignature,
		minerSignature:     result.MinerSignature,
	})
	wm.node.PostProof(proof)
	log.WithFields(log.Fields{
		"task-id": result.TaskId,
	}).Debug("send proof to kafka")
	return proof, nil
}

func (wm *WorkerManager) standardTaskResult(worker *Worker, task *odysseus.TaskContent, result *omanager.SubmitTaskResult) (*odysseus.TaskProof, error) {
	if worker.info.nodeInfo == nil {
		return nil, errors.New("unknown worker node info")
	}
	//log.WithFields(log.Fields{
	//	"task-id":               task.TaskId,
	//	"task-type":             task.TaskType,
	//	"task-kind":             task.TaskKind,
	//	"task-success":          result.IsSuccessed,
	//	"task-execute-duration": result.TaskExecuteDuration,
	//}).Debug("got task result")

	if result.IsSuccessed == false {
		receipt := wm.makeReceipt(worker, task, result, ErrExecuteFailed)
		wm.node.PostResult(receipt)
		return nil, nil
	}

	stdlib := wm.std.GetStdLib(task.TaskType)
	if stdlib == nil {
		log.WithField("task-id", task.TaskId).Error("not found stdlib to verify")
		return nil, errors.New("not found stdlib to verify")
	}

	if stdlib.VerifyResult(string(task.TaskParam), result.TaskResultBody) == true {
		log.WithField("task-id", task.TaskId).Debug("stdlib to verify passed")
	} else {
		log.WithField("task-id", task.TaskId).Debug("stdlib to verify failed")
		return nil, errors.New("stdlib to verify failed")
	}

	paramHash := crypto.Keccak256Hash(task.TaskParam)
	resultHash := crypto.Keccak256Hash(result.TaskResultBody)
	{
		// verify container_signature and miner_signature
		// container_signature = sign(hash(task_id+hash(task_param)+hash(task_result)))
		verified := utils.VerifySignature(utils.CombineBytes([]byte(result.TaskId), paramHash[:], resultHash[:]),
			result.ContainerSignature, task.ContainerPubkey)
		log.WithFields(log.Fields{
			"task-id":                  task.TaskId,
			"containerSignatureVerify": verified,
			"taskkind":                 task.TaskKind,
		}).Debug("container signature verify")
		if !verified {
			// todo: handle signature verify failed
		}
	}
	{
		// verify miner_signature
		// miner_signature = sign(hash((task_id+hash(task_param)+hash(task_result)))
		verified := utils.VerifySignature(utils.CombineBytes([]byte(result.TaskId), paramHash[:], resultHash[:]),
			result.MinerSignature, utils.FromHex(worker.info.nodeInfo.MinerPubkey))
		log.WithFields(log.Fields{
			"task-id":              task.TaskId,
			"minerSignatureVerify": verified,
			"taskkind":             task.TaskKind,
		}).Debug("miner signature verify")
		if !verified {
			// todo: handle signature verify failed
		}
	}

	receipt := wm.makeReceipt(worker, task, result, Succeed)
	wm.node.PostResult(receipt)

	now := time.Now().Unix()
	//manager_signature = sign(hash((task_id+hash(task_param)+hash(task_result)+container_signature+miner_signature+workload+time))
	dataHash := crypto.Keccak256Hash(utils.CombineBytes([]byte(result.TaskId), paramHash[:], resultHash[:],
		worker.ProfitAccount().Bytes(), worker.WorkerAccount().Bytes(), result.ContainerSignature, result.MinerSignature, big.NewInt(int64(task.TaskWorkload)).Bytes()),
		big.NewInt(now).Bytes())

	signature, err := wm.node.Sign(dataHash[:])
	if err != nil {
		log.WithError(err).Error("sign result failed")
		return nil, err
	}
	proof := wm.makeTaskProof(worker, task, taskProof{
		paramHash:          paramHash.Bytes(),
		resultHash:         resultHash.Bytes(),
		finishTime:         uint64(now),
		nmSignature:        signature,
		containerSignature: result.ContainerSignature,
		minerSignature:     result.MinerSignature,
	})
	wm.node.PostProof(proof)

	log.WithFields(log.Fields{
		"task-id": result.TaskId,
	}).Debug("send proof to kafka")
	return proof, nil
}
