package server

import (
	"errors"
	"github.com/ethereum/go-ethereum/common"
	lru "github.com/hashicorp/golang-lru"
	omanager "github.com/odysseus/odysseus-protocol/gen/proto/go/nodemanager/v1"
	log "github.com/sirupsen/logrus"
	"time"
)

type workerInfo struct {
	nodeInfo         *omanager.NodeInfoResponse
	deviceUsageInfo  []*omanager.DeviceUsage
	deviceInfo       *omanager.DeviceInfoMessage
	deviceStatusInfo *omanager.StatusResponse
	resourceInfo     *omanager.SubmitResourceMap
}

type sendMsgCallback struct {
	msg      *omanager.ManagerMessage
	callback func(err error) bool
}
type Worker struct {
	quit     chan interface{}
	taskCh   chan *dispatchTask
	msgCh    chan *omanager.WorkerMessage
	sendCh   chan sendMsgCallback
	resultCh chan *omanager.SubmitTaskResult

	uuid            int64 // worker uuid in the local.
	registed        bool  // worker is registed to this nm.
	online          bool
	disconnect      bool
	nonce           int
	latestNmValue   string
	addFirstSucceed bool

	info           workerInfo
	workerAddr     string // worker address from public-key
	deviceInfoHash []byte
	recentTask     *lru.ARCCache
	status         string
	errCh          chan error

	stream omanager.NodeManagerService_RegisterWorkerServer
}

func (w *Worker) ProfitAccount() common.Address {
	if w.info.nodeInfo != nil {
		return common.HexToAddress(w.info.nodeInfo.BenefitAddress)
	}
	return common.Address{}
}

func (w *Worker) WorkerAccount() common.Address {
	return common.HexToAddress(w.workerAddr)
}

func (w *Worker) Disconnect() {
	w.disconnect = true
}

func (w *Worker) DisConnected() bool {
	return w.disconnect
}

func (w *Worker) SendToWorker(msg *omanager.ManagerMessage, callback func(err error) bool) {
	w.sendCh <- sendMsgCallback{msg, callback}
}

func (w *Worker) SendMessage() {
	l := log.WithField("worker-uuid", w.uuid)
	defer func() {
		if e := recover(); e != nil {
			l.WithField("worker-addr", w.workerAddr).Error("worker handle message panic")
		}
	}()
	for w.DisConnected() == false {
		select {
		case <-w.quit:
			return
		case m := <-w.sendCh:
			if w.DisConnected() {
				return
			}
			start := time.Now()
			if m.msg.Message != nil {
				err := w.stream.Send(m.msg)
				if err != nil {
					log.WithError(err).Error("send message to worker failed")
				}
				if m.callback != nil {
					m.callback(err)
				}
			}
			log.WithField("worker-addr", w.workerAddr).WithField("send duration", time.Now().Sub(start).String()).Debug("send msg to worker")
		}
	}

}

func (w *Worker) RecvMessage() {
	l := log.WithField("worker-uuid", w.uuid)
	defer func() {
		if e := recover(); e != nil {
			l.WithField("worker-addr", w.workerAddr).Error("worker handle message panic")
		}
	}()
	for {
		start := time.Now()
		wmsg, err := w.stream.Recv()
		if err != nil {
			l.WithError(err).WithField("worker-addr", w.workerAddr).WithField("recv duration", time.Now().Sub(start).String()).Error("recv msg failed")
			w.errCh <- errors.New("recv msg failed")
			return
		}
		if w.DisConnected() {
			return
		}
		w.msgCh <- wmsg
	}
}
