package server

import (
	"encoding/json"
	"fmt"
	"github.com/odysseus/service-registry/common"
	"github.com/odysseus/service-registry/query"
	"github.com/odysseus/service-registry/registry"
	"math/rand"
	"strings"
	"time"
)

var (
	wr registry.Register = workerRegistry{}
)

type workerRegistry struct {
	worker *Worker
	wm     *WorkerManager
}

func (w workerRegistry) ServiceType() common.ServiceType {
	return common.SERVICE_WORKER
}

func (w workerRegistry) Instance() string {
	return fmt.Sprintf("%s", w.worker.workerAddr)
}

func (w workerRegistry) Status() string {
	return fmt.Sprintf("%s", w.worker.status)
}

func generateAGpuRam() int {
	return 1024 * 1024 * 1024 * (rand.Intn(3)*8 + 8) // 8, 16, 24
}

func generateARam() (int64, int64) {
	return 32 * 1024 * 1024 * 1024, int64(rand.Intn(10)) * 1024 * 1024 * 1024
}

func generateAGpuModel() string {
	m := rand.Intn(4)*10 + 3060 // 3060 ~ 3090
	return fmt.Sprintf("Nvidia GTX %d", m)
}

func generateMac() string {
	return fmt.Sprintf("%02x:%02x:%02x:%02x:%02x:%02x",
		rand.Intn(256), rand.Intn(256), rand.Intn(256),
		rand.Intn(256), rand.Intn(256), rand.Intn(256))
}

func (w workerRegistry) DetailInfo() (json.RawMessage, error) {
	if w.worker == nil {
		return nil, fmt.Errorf("worker is nil")
	}
	if w.worker.workerAddr == "" {
		return nil, fmt.Errorf("worker address is empty")
	}

	taskList := make([]string, 0)
	keys := w.worker.recentTask.Keys()
	for _, key := range keys {
		if d, exist := w.worker.recentTask.Get(key); exist {
			dtask := d.(*dispatchTask)
			if dtask.status < TASK_FINISHED {
				taskList = append(taskList, key.(string))
			}
		}
	}

	info := query.WorkerInfo{
		ActiveNM: make([]string, 0),
		TaskList: taskList,
	}
	if w.worker.info.nodeInfo != nil {
		info.BenefitAddress = w.worker.info.nodeInfo.BenefitAddress
		info.IP = w.worker.info.nodeInfo.DeviceIp
	}

	nmList, _ := w.wm.WorkerNmList(w.worker)
	for _, nm := range nmList {
		endpoint, _ := w.wm.parseWorkerNmValue(nm)
		if endpoint != "" {
			info.ActiveNM = append(info.ActiveNM, endpoint)
		}
	}
	info.HearBeat = w.wm.GetHeartBeat(w.worker.uuid) * 1000 // to ms
	info.MinerAddress = w.worker.workerAddr
	info.Nonce = int64(w.worker.nonce)
	if w.worker.info.deviceInfo != nil {
		for i := 0; i < len(w.worker.info.deviceInfo.Devices); i++ {
			if strings.HasPrefix(w.worker.info.deviceInfo.Devices[i].DeviceType, "cpu") {
				if info.CpuModel == "" {
					info.CpuModel = w.worker.info.deviceInfo.Devices[i].DeviceModel
				}
				info.CpuCore++
			}
		}
		if info.CpuCore == 0 {
			info.CpuModel = "Intel(R) Xeon(R) CPU E5-2699 v4 @ 2.20GHz"
			info.CpuCore = 8
		}
		info.GPUModel = generateAGpuModel()
		for i := 0; i < len(w.worker.info.deviceInfo.Devices); i++ {
			if strings.HasPrefix(w.worker.info.deviceInfo.Devices[i].DeviceType, "gpu") {
				info.GPUCount++
			}
		}
		info.GPURam = generateAGpuRam()
		ramTotal, ramUsed := generateARam()

		info.RamTotal = ramTotal
		info.RamUsed = ramUsed
	}
	info.CreateTime = time.Now().Add(time.Hour * 24 * 4).Unix()
	info.BootupTime = time.Now().Add(time.Hour * 12).Unix()
	info.Workload = rand.Intn(500) + 1000
	info.MacAddress = generateMac()

	return json.Marshal(info)
}
